import numpy as np

def get_indices(X_shape, HF, WF, stride, pad):
    """
        Returns index matrices in order to transform our input image into a matrix.
        Parameters:
        -X_shape: Input image shape.
        -HF: filter height.
        -WF: filter width.
        -stride: stride value.
        -pad: padding value.
        Returns:
        -i: matrix of index i.
        -j: matrix of index j.
        -d: matrix of index d. 
            (Use to mark delimitation for each channel
            during multi-dimensional arrays indexing).
    """
    # get input size
    m, n_C, n_H, n_W = X_shape

    # get output size
    out_h = int((n_H + 2 * pad - HF) / stride) + 1
    out_w = int((n_W + 2 * pad - WF) / stride) + 1
  
    # ----Compute matrix of index i----

    # Level 1 vector.
    level1 = np.repeat(np.arange(HF), WF)
    # Duplicate for the other channels.
    level1 = np.tile(level1, n_C)
    # Create a vector with an increase by 1 at each level.
    everyLevels = stride * np.repeat(np.arange(out_h), out_w)
    # Create matrix of index i at every levels for each channel.
    i = level1.reshape(-1, 1) + everyLevels.reshape(1, -1)

    # ----Compute matrix of index j----
    
    # Slide 1 vector.
    slide1 = np.tile(np.arange(WF), HF)
    # Duplicate for the other channels.
    slide1 = np.tile(slide1, n_C)
    # Create a vector with an increase by 1 at each slide.
    everySlides = stride * np.tile(np.arange(out_w), out_h)
    # Create matrix of index j at every slides for each channel.
    j = slide1.reshape(-1, 1) + everySlides.reshape(1, -1)

    # ----Compute matrix of index d----

    # This is to mark delimitation for each channel
    # during multi-dimensional arrays indexing.
    d = np.repeat(np.arange(n_C), HF * WF).reshape(-1, 1)

    return i, j, d

def im2col(X, HF, WF, stride, pad):
    """
        Transforms our input image into a matrix.
        Parameters:
        - X: input image.
        - HF: filter height.
        - WF: filter width.
        - stride: stride value.
        - pad: padding value.
        Returns:
        -cols: output matrix.
    """
    # Padding
    X_padded = np.pad(X, ((0,0), (0,0), (pad, pad), (pad, pad)), mode='constant')
    i, j, d = get_indices(X.shape, HF, WF, stride, pad)
    # Multi-dimensional arrays indexing.
    cols = X_padded[:, d, i, j]
    cols = np.concatenate(cols, axis=-1)
    return cols

def col2im(dX_col, X_shape, HF, WF, stride, pad):
    """
        Transform our matrix back to the input image.
        Parameters:
        - dX_col: matrix with error.
        - X_shape: input image shape.
        - HF: filter height.
        - WF: filter width.
        - stride: stride value.
        - pad: padding value.
        Returns:
        -x_padded: input image with error.
    """
    # Get input size
    N, D, H, W = X_shape
    # Add padding if needed.
    H_padded, W_padded = H + 2 * pad, W + 2 * pad
    X_padded = np.zeros((N, D, H_padded, W_padded))
    
    # Index matrices, necessary to transform our input image into a matrix. 
    i, j, d = get_indices(X_shape, HF, WF, stride, pad)
    # Retrieve batch dimension by spliting dX_col N times: (X, Y) => (N, X, Y)
    dX_col_reshaped = np.array(np.hsplit(dX_col, N))
    # Reshape our matrix back to image.
    # slice(None) is used to produce the [::] effect which means "for every elements".
    np.add.at(X_padded, (slice(None), d, i, j), dX_col_reshaped)
    # Remove padding from new image if needed.
    if pad == 0:
        return X_padded
    elif type(pad) is int:
        return X_padded[pad:-pad, pad:-pad, :, :]

class AvgPool():
    
    def __init__(self, filter_size, stride=1, padding=0):
        self.f = filter_size
        self.s = stride
        self.p = padding
        self.cache = None

    def forward(self, X):
        """
            Apply average pooling.
            Parameters:
            - X: Output of activation function.
            
            Returns:
            - A_pool: X after average pooling layer. 
        """
        self.cache = X

        m, n_C_prev, n_H_prev, n_W_prev = X.shape
        n_C = n_C_prev
        n_H = int((n_H_prev + 2 * self.p - self.f)/ self.s) + 1
        n_W = int((n_W_prev + 2 * self.p - self.f)/ self.s) + 1
        
        X_col = im2col(X, self.f, self.f, self.s, self.p)
        X_col = X_col.reshape(n_C, X_col.shape[0]//n_C, -1)
        A_pool = np.mean(X_col, axis=1)
        # Reshape A_pool properly.
        A_pool = np.array(np.hsplit(A_pool, m))
        A_pool = A_pool.reshape(m, n_C, n_H, n_W)

        return A_pool

    def backward(self, dout):
        """
            Distributes error through pooling layer.
            Parameters:
            - dout: Previous layer with the error.
            
            Returns:
            - dX: Conv layer updated with error.
        """
        X = self.cache
        m, n_C_prev, n_H_prev, n_W_prev = X.shape

        n_C = n_C_prev
        n_H = int((n_H_prev + 2 * self.p - self.f)/ self.s) + 1
        n_W = int((n_W_prev + 2 * self.p - self.f)/ self.s) + 1

        dout_flatten = dout.reshape(n_C, -1) / (self.f * self.f)
        dX_col = np.repeat(dout_flatten, self.f*self.f, axis=0)
        dX = col2im(dX_col, X.shape, self.f, self.f, self.s, self.p)
        # Reshape dX properly.
        dX = dX.reshape(m, -1)
        dX = np.array(np.hsplit(dX, n_C_prev))
        dX = dX.reshape(m, n_C_prev, n_H_prev, n_W_prev)
        return dX












from Terminatetensorflow.conv import *
from Terminatetensorflow.pool import *
from Terminatetensorflow.dense import *
from Terminatetensorflow.actvfn import *
from Terminatetensorflow.loss import *


class LeNet():
    def __init__(self):
        self.conv1 = Conv(nb_filters = 6, filter_size = 5, nb_channels = 1)
        self.pool1 = AvgPool(filter_size = 2, stride = 2)
        self.conv2 = Conv(nb_filters = 16, filter_size = 5, nb_channels = 6)
        self.pool2 = AvgPool(filter_size = 2, stride = 2)
        self.pool2_shape = None
        self.fc1 = layer(128, 400, "tanh")
        self.fc2 = layer(84, 128, "tanh")
        self.fc3 = layer(10,84,"softmax")
        self.A = []
        self.actCache = []
        self.layers = [self.conv1, self.conv2, self.fc1, self.fc2, self.fc3]
        
        
    def forward(self, X):
        self.A = []
        conv1 = self.conv1.forward(X) #(6x28x28)
        self.actCache.append(conv1)
        act1 = tanh(conv1)
        pool1 = self.pool1.forward(act1) #(6x14x14)

        conv2 = self.conv2.forward(pool1) #(16x10x10)
        self.actCache.append(conv2)
        act2 = tanh(conv2)
        pool2 = self.pool2.forward(act2) #(16x5x5)
        
        self.pool2_shape = pool2.shape #Need it in backpropagation.
        pool2_flatten = pool2.reshape(self.pool2_shape[0], -1) #(1x400)
        self.A.append(pool2_flatten.T)
        fc1 = self.fc1.forward_prop(pool2_flatten.T) #(1x120) 
        self.A.append(fc1)
        fc2 = self.fc2.forward_prop(fc1) #(1x84)    
        
        # print(fc2.shape)
        self.A.append(fc2)
        fc3 = self.fc3.forward_prop(fc2) #(1x10)
        self.A.append(fc3)

        y_pred = fc3

        return y_pred
    
    def backward(self, y_pred, y):
        deltaL = y_pred - y
        #Compute gradient for weight/bias between fc3 and fc2.
        deltaL, dW5, db5= self.fc3.backward_prop(self.A[2],0,y,output_layer = True)
        #Compute error at fc2 layer.
        #deltaL = self.tanh4.backward(deltaL) #(1x84) 
        
        #Compute gradient for weight/bias between fc2 and fc1.
        deltaL, dW4, db4= self.fc2.backward_prop(self.A[1],deltaL,y,output_layer = False)
        #Compute error at fc1 layer.
       # deltaL = self.tanh3.backward(deltaL) #(1x120)
        
        #Compute gradient for weight/bias between fc1 and pool2 and compute 
        #error too (don't need to backpropagate through tanh here).
        deltaL, dW3, db3 = self.fc1.backward_prop(self.A[0],deltaL,y,output_layer = False) #(1x400)
        deltaL = deltaL.reshape(self.pool2_shape) #(16x5x5)
        
        
        #Distribute error through pool2 to conv2.
        deltaL = self.pool2.backward(deltaL) #(16x10x10)
        #Distribute error through tanh.
        deltaL = derivative_tanh(self.actCache[1]) * deltaL
        
        #Compute gradient for weight/bias at conv2 layer and backpropagate
        #error to conv1 layer.
        deltaL, dW2, db2 = self.conv2.backward(deltaL) #(6x14x14)

        #Distribute error through pool1 by creating a temporary pooling layer
        #of conv1 shape.
        deltaL = self.pool1.backward(deltaL) #(6x28x28)
        #Distribute error through tanh.
        deltaL = derivative_tanh(self.actCache[0]) * deltaL
    
        #Compute gradient for weight/bias at conv1 layer and backpropagate
        #error at conv1 layer.
        deltaL, dW1, db1 = self.conv1.backward(deltaL) #(1x32x32)
    
        grads = { 
                'dW1': dW1, 'db1': db1,
                'dW2': dW2, 'db2': db2, 
                'dW3': dW3, 'db3': db3,
                'dW4': dW4, 'db4': db4,
                'dW5': dW5, 'db5': db5
        }

        return grads
    
    
    def get_params(self):
        params = {}
        for i, layer in enumerate(self.layers):
            try:
                params['W' + str(i+1)] = layer.W['val']
                params['b' + str(i+1)] = layer.b['val']
            except:
                  params['W' + str(i+1)] = layer.weights
                  params['b' + str(i+1)] = layer.biases            

        return params

    def set_params(self, params):
        for i, layer in enumerate(self.layers):
            try:
                layer.W['val'] = params['W'+ str(i+1)]
                layer.b['val'] = params['b' + str(i+1)]
            except:
                layer.weights = params['W'+ str(i+1)]
                layer.biases = params['b' + str(i+1)]
                
                
    def update_params(self,grads,learning_rate):
        for i, layer in enumerate(self.layers):       
            try:                
                layer.W['val'] = layer.W['val'] - learning_rate*grads['dW'+str(i+1)] 
                layer.b['val'] = layer.b['val'] - learning_rate*grads['db'+str(i+1)]
            except:
                    layer.weights = layer.weights - learning_rate*grads['dW'+str(i+1)]    
                    layer.biases = layer.biases - learning_rate*grads['db'+str(i+1)] 
                    
    def update_parameters_momentum(self, grads, learning_rate, gamma=0.9):
        
        for i, layer in enumerate(self.layers): 
            try:     
            
               layer.v_w = gamma * layer.v_w +  (1 - gamma)* grads['dW'+str(i+1)]
               layer.v_b = gamma * layer.v_b +  (1 - gamma)* grads['db'+str(i+1)]
               layer.W['val'] = layer.W['val'] - learning_rate * layer.v_w
               layer.b['val'] = layer.b['val'] - learning_rate * layer.v_b
                           
            except:
                
                layer.v_w = gamma*layer.v_w +  (1 - gamma)* layer.dW
                layer.v_b = gamma*layer.v_b +  (1 - gamma)* layer.db
                layer.weights = layer.weights - learning_rate * layer.v_w
                layer.biases = layer.biases -  learning_rate  *  layer.v_b
                
    def update_parameters_rmsprop(self, grads, learning_rate, beta=0.9):
        
        for i, layer in enumerate(self.layers): 
            try:     
            
               layer.s_w = beta * layer.s_w +  (1 - beta)* np.power(grads['dW'+str(i+1)],2)
               layer.s_b = beta * layer.s_b +  (1 - beta)* np.power(grads['db'+str(i+1)],2)
               layer.W['val'] = layer.W['val'] - learning_rate * np.divide(grads['dW'+str(i+1)],np.sqrt(layer.s_w))
               layer.b['val'] = layer.b['val'] - learning_rate * np.divide(grads['db'+str(i+1)],np.sqrt(layer.s_b))
                           
            except:
                
                layer.s_w = beta*layer.s_w +  (1 - beta)* np.power(grads['dW'+str(i+1)],2)
                layer.s_b = beta*layer.s_b +  (1 - beta)* np.power(grads['db'+str(i+1)],2)
                layer.weights = layer.weights - learning_rate * np.divide(grads['dW'+str(i+1)],np.sqrt(layer.s_w))
                layer.biases = layer.biases -  learning_rate  *  np.divide(grads['db'+str(i+1)],np.sqrt(layer.s_b))

                
    def update_parameters_adam(self, grads, learning_rate = 0.01, t = 2, beta1 = 0.9, beta2 = 0.999,  epsilon = 1e-8):
        for i, layer in enumerate(self.layers): 
            try:     
               layer.v_w = beta1 * layer.v_w +  (1 - beta1)* np.atleast_2d(grads['dW'+str(i+1)])
               layer.v_b = beta1 * layer.v_b +  (1 - beta1)* grads['db'+str(i+1)]
               v_wcorrected = layer.v_w / (1-np.power(beta1,t))
               v_bcorrected = layer.v_b / (1-np.power(beta1,t))
               layer.s_w = beta2*layer.s_w + (1-beta2)*np.power(grads['dW'+str(i+1)],2)
               layer.s_b = beta2*layer.s_b + (1-beta2)*np.power(grads['db'+str(i+1)],2)
               s_wcorrected = layer.s_w / (1-np.power(beta2,t))
               s_bcorrected = layer.s_b / (1-np.power(beta2,t))
               layer.W['val'] = layer.W['val'] - learning_rate*v_wcorrected / np.sqrt(s_wcorrected + epsilon)
               layer.b['val'] = layer.b['val'] - learning_rate*v_bcorrected / np.sqrt(s_bcorrected + epsilon)

            except:
                layer.v_w = beta1*layer.v_w +  (1 - beta1)* layer.dW
                layer.v_b = beta1*layer.v_b +  (1 - beta1)* layer.db
                v_wcorrected = layer.v_w / (1-np.power(beta1,t))
                v_bcorrected = layer.v_b / (1-np.power(beta1,t))
                layer.s_w = beta2*layer.s_w + (1-beta2)*np.power(layer.dW,2)
                layer.s_b = beta2*layer.s_b + (1-beta2)*np.power(layer.db,2)
                s_wcorrected = layer.s_w / (1-np.power(beta2,t))
                s_bcorrected = layer.s_b / (1-np.power(beta2,t))
                layer.weights = layer.weights - learning_rate*v_wcorrected / np.sqrt(s_wcorrected + epsilon)
                layer.biases = layer.biases - learning_rate*v_bcorrected / np.sqrt(s_bcorrected + epsilon)
                
                

    def predict(self, X):
        output = self.forward(X)

        _pred = np.array(output[0])        
        pred = np.zeros((10,X.shape[1]))  
        pred = np.argmax(output,axis=0)
        
        return np.atleast_2d(pred)

    def fit(self, X, Y, epochs, learning_rate, costt, batch_size, optimizer="adam"):
        Y_class = np.zeros((10,Y.shape[1]))  
        Y_class = np.atleast_2d(np.argmax(Y,axis=0))
                
        batch_number = int(-(-X.shape[0]/batch_size // 1))
        metrics_output = {}
        
        for i in range(epochs):
            
            current_batch=0
            print('\repoch:{}/{} [{}] {}%'.format(i+1, epochs, '.' * (50), 0), end='\r')
            for j in range(batch_number-1):

                current_batch += 1
                if j == batch_number-1:
                    
                    x = X[batch_size*(j):,:,:,:]
                    y = Y[:,batch_size*(j):]
                    
                else:
                    x = X[batch_size*(j):batch_size*(j+1),:,:,:]
                    y = Y[:,batch_size*(j):batch_size*(j+1)]
                    
                y_hat = self.forward(x)
                
                # acc = accuracy_metrics(y_hat.T, y.T) * 100
                
                
                if (costt == "multiclass"):
                    loss = compute_multiclass_loss(y, y_hat)
                elif (costt == "mse"):
                    loss = compute_mse_loss(y, y_hat)
                    
                grads = self.backward(y_hat, y)
                if optimizer == "adam":
                    self.update_parameters_adam(grads, learning_rate)
                    
                elif optimizer == "rmsprop":
                    self.update_parameters_rmsprop(grads, learning_rate)
                
                elif optimizer == "momentum":
                    self.update_parameters_momentum(grads, learning_rate)
                
                else:
                    self.update_params(grads, learning_rate)

                done = int(100*current_batch/batch_number)
                print('\repoch:{}/{} [{}{}] {}% '.format(i+1, epochs,'█' * int(done/2), '.' * int(50-int(done/2)), done, ), end='\r')
            
            print("Epoch", i+1, "->")
            print("\t\tLoss =", ("%.4f" % loss ))
            metrics_output["loss"+str(i)] = loss
            # print("\t\Acc =", ("%.4f" % acc ))
            
        return metrics_output
        





import numpy as np
from Terminatetensorflow.actvfn import *

def get_indices(X_shape, HF, WF, stride, pad):
    """
        Returns index matrices in order to transform our input image into a matrix.
        Parameters:
        -X_shape: Input image shape.
        -HF: filter height.
        -WF: filter width.
        -stride: stride value.
        -pad: padding value.
        Returns:
        -i: matrix of index i.
        -j: matrix of index j.
        -d: matrix of index d. 
            (Use to mark delimitation for each channel
            during multi-dimensional arrays indexing).
    """
    # get input size
    m, n_C, n_H, n_W = X_shape

    # get output size
    out_h = int((n_H + 2 * pad - HF) / stride) + 1
    out_w = int((n_W + 2 * pad - WF) / stride) + 1
  
    # ----Compute matrix of index i----

    # Level 1 vector.
    level1 = np.repeat(np.arange(HF), WF)
    # Duplicate for the other channels.
    level1 = np.tile(level1, n_C)
    # Create a vector with an increase by 1 at each level.
    everyLevels = stride * np.repeat(np.arange(out_h), out_w)
    # Create matrix of index i at every levels for each channel.
    i = level1.reshape(-1, 1) + everyLevels.reshape(1, -1)

    # ----Compute matrix of index j----
    
    # Slide 1 vector.
    slide1 = np.tile(np.arange(WF), HF)
    # Duplicate for the other channels.
    slide1 = np.tile(slide1, n_C)
    # Create a vector with an increase by 1 at each slide.
    everySlides = stride * np.tile(np.arange(out_w), out_h)
    # Create matrix of index j at every slides for each channel.
    j = slide1.reshape(-1, 1) + everySlides.reshape(1, -1)

    # ----Compute matrix of index d----

    # This is to mark delimitation for each channel
    # during multi-dimensional arrays indexing.
    d = np.repeat(np.arange(n_C), HF * WF).reshape(-1, 1)

    return i, j, d

def im2col(X, HF, WF, stride, pad):
    """
        Transforms our input image into a matrix.
        Parameters:
        - X: input image.
        - HF: filter height.
        - WF: filter width.
        - stride: stride value.
        - pad: padding value.
        Returns:
        -cols: output matrix.
    """
    # Padding
    X_padded = np.pad(X, ((0,0), (0,0), (pad, pad), (pad, pad)), mode='constant')
    i, j, d = get_indices(X.shape, HF, WF, stride, pad)
    # Multi-dimensional arrays indexing.
    cols = X_padded[:, d, i, j]
    cols = np.concatenate(cols, axis=-1)
    return cols

def col2im(dX_col, X_shape, HF, WF, stride, pad):
    """
        Transform our matrix back to the input image.
        Parameters:
        - dX_col: matrix with error.
        - X_shape: input image shape.
        - HF: filter height.
        - WF: filter width.
        - stride: stride value.
        - pad: padding value.
        Returns:
        -x_padded: input image with error.
    """
    # Get input size
    N, D, H, W = X_shape
    # Add padding if needed.
    H_padded, W_padded = H + 2 * pad, W + 2 * pad
    X_padded = np.zeros((N, D, H_padded, W_padded))
    
    # Index matrices, necessary to transform our input image into a matrix. 
    i, j, d = get_indices(X_shape, HF, WF, stride, pad)
    # Retrieve batch dimension by spliting dX_col N times: (X, Y) => (N, X, Y)
    dX_col_reshaped = np.array(np.hsplit(dX_col, N))
    # Reshape our matrix back to image.
    # slice(None) is used to produce the [::] effect which means "for every elements".
    np.add.at(X_padded, (slice(None), d, i, j), dX_col_reshaped)
    # Remove padding from new image if needed.
    if pad == 0:
        return X_padded
    elif type(pad) is int:
        return X_padded[pad:-pad, pad:-pad, :, :]

class Conv():
    
    def __init__(self, nb_filters, filter_size, nb_channels, stride=1, padding=0):
        self.n_F = nb_filters
        self.f = filter_size
        self.n_C = nb_channels
        self.s = stride
        self.p = padding
        # Xavier-Glorot initialization - used for sigmoid, tanh.
        self.W = {'val': np.random.randn(self.n_F, self.n_C, self.f, self.f)*np.sqrt(1./self.f) ,
                  'grad': np.zeros((self.n_F, self.n_C, self.f, self.f))}  
        self.b = {'val': np.random.randn(self.n_F), 'grad': np.zeros((self.n_F))}
        
        self.v_w = np.zeros((self.n_F, self.n_C, self.f, self.f)) 
        self.v_b = np.zeros((self.n_F))
        self.s_w = np.zeros((self.n_F, self.n_C, self.f, self.f)) 
        self.s_b = np.zeros((self.n_F))
        self.w_squared = np.zeros((self.n_F, self.n_C, self.f, self.f)) 
        
        self.cache = None

    def conv_single_step(self, X_slice_prev, W, b):
        s = np.multiply(X_slice_prev, W) + float(b)
        # Sum over all entries of the volume s.
        Z = np.sum(s)
        
        return Z
    
    def zero_pad(self, X, P):
        return np.pad(X, ((0,0), (P,P),(P,P), (0,0)), mode='constant', constant_values = (0,0))
    
    def fprop_slow(self,X):
        X =  self.zero_pad(X , self.padding)
        
        m = X.shape[0]    
        FH = self.filters.shape[0]
        FW = self.filters.shape[1]
        
        # output array
        Z = np.zeros((m, self.n_H, self.n_W, self.n_C)) 
        
        for i in range(m):                      # loop over the batch of training examples
            X_prev = X[i,:,:,:]               # Select ith training example's padded activation
            for h in range(self.n_H):           # loop over vertical axis of the output volume
                # Find the vertical start and end of the current "slice" 
                vert_start = h * self.stride
                vert_end = vert_start + FH
                for w in range(self.n_W):       # loop over horizontal axis of the output volume
                    # Find the horizontal start and end of the current "slice"
                    horiz_start = w * self.stride
                    horiz_end = horiz_start + FW
                    for c in range(self.n_C):   # loop over channels (= #filters) of the output volume                      
                        X_slice_prev = X_prev[vert_start:vert_end,horiz_start:horiz_end,:]
                        # Convolve the (3D) slice with the correct filter W and bias b, to get back one output neuron.
                        Z[i, h, w, c] = self.conv_single_step(X_slice_prev, self.filters[:,:,:,c], self.b[:,:,:,c])
        
        if self.act_fn == "sigmoid":
            A = sigmoid(Z)
        elif self.act_fn == "relu":
            A = relu(Z)
        elif self.act_fn == "tanh":
            A = tanh(Z)
        elif self.act_fn == "linear":
            A = linear(Z)
        elif self.act_fn == "softmax":
            A = softmax(Z)
            
        self.cache = X
        
        return out    

    def forward(self, X):

        m, n_C_prev, n_H_prev, n_W_prev = X.shape

        n_C = self.n_F
        n_H = int((n_H_prev + 2 * self.p - self.f)/ self.s) + 1
        n_W = int((n_W_prev + 2 * self.p - self.f)/ self.s) + 1
        
        X_col = im2col(X, self.f, self.f, self.s, self.p)
        w_col = self.W['val'].reshape((self.n_F, -1))
        b_col = self.b['val'].reshape(-1, 1)
        # Perform matrix multiplication.
        out = w_col @ X_col + b_col
        # Reshape back matrix to image.
        out = np.array(np.hsplit(out, m)).reshape((m, n_C, n_H, n_W))
        self.cache = X, X_col, w_col
        return out

    def backward(self, dout):

        X, X_col, w_col = self.cache
        m, _, _, _ = X.shape
        # Compute bias gradient.
        self.b['grad'] = np.sum(dout, axis=(0,2,3)) 
        # Reshape dout properly.
        dout = dout.reshape(dout.shape[0] * dout.shape[1], dout.shape[2] * dout.shape[3])
        dout = np.array(np.vsplit(dout, m))
        dout = np.concatenate(dout, axis=-1)
        # Perform matrix multiplication between reshaped dout and w_col to get dX_col.
        dX_col = w_col.T @ dout
        # Perform matrix multiplication between reshaped dout and X_col to get dW_col.
        dw_col = dout @ X_col.T
        # Reshape back to image (col2im).
        dX = col2im(dX_col, X.shape, self.f, self.f, self.s, self.p)
        # Reshape dw_col into dw.
        self.W['grad'] = dw_col.reshape((dw_col.shape[0], self.n_C, self.f, self.f)) 
                
        return dX, self.W['grad'], self.b['grad']

    def bprop_slow(self, dZ):
        (m, n_H_prev, n_W_prev, n_C_prev) = self.A_prev.shape
        
        dA_prev = np.zeros((m, n_H_prev, n_W_prev, n_C_prev))                           
        self.dW = np.zeros((self.f, self.f, n_C_prev, self.n_C))
        self.db = np.zeros((1, 1, 1, self.n_C))
        
        A_prev_pad = np.pad(self.A_prev, ((0, 0), (self.pad, self.pad), (self.pad, self.pad), (0, 0)), 'constant', constant_values=0)
        dA_prev_pad = np.pad(dA_prev, ((0, 0), (self.pad, self.pad), (self.pad, self.pad), (0, 0)), 'constant', constant_values=0)
    
        if self.act_fn == "sigmoid":
            dZ = derivative_sigmoid(dZ)
        elif self.act_fn == "tanh":
            dZ = derivative_tanh(dZ)
        elif self.act_fn == "relu":
            dZ = d_relu(dZ)
        elif self.act_fn == "linear":
            dZ = d_linear(dZ)    
    
        for i in range(m):                       # loop over the training examples
            
            # select ith training example from A_prev_pad and dA_prev_pad
            a_prev_pad = A_prev_pad[i]
            da_prev_pad = dA_prev_pad[i]
            
            for h in range(self.n_H):                   # loop over vertical axis of the output volume
                for w in range(self.n_W):               # loop over horizontal axis of the output volume
                    for c in range(self.n_C):           # loop over the channels of the output volume
                    
                        # Find the corners of the current "slice"
                        vert_start = h * self.stride
    
                        vert_end = vert_start + self.f
                        horiz_start = w * self.stride
    
                        horiz_end = horiz_start + self.f
                        
                        # Use the corners to define the slice from a_prev_pad
                        a_slice = a_prev_pad[vert_start:vert_end, horiz_start:horiz_end, :]
    
                        # Update gradients for the window and the filter's parameters using the code formulas given above
                        da_prev_pad[vert_start:vert_end, horiz_start:horiz_end, :] += self.W[:,:,:,c] * dZ[i, h, w, c]
                        self.dW[:,:,:,c] += a_slice * dZ[i, h, w, c]
                        self.db[:,:,:,c] += dZ[i, h, w, c]
                        
            dA_prev[i, :, :, :] = da_prev_pad[self.pad:-self.pad, self.pad:-self.pad, :]
        

        assert(dA_prev.shape == (m, n_H_prev, n_W_prev, n_C_prev))
        
        return dA_prev



import matplotlib.pyplot as plt
import matplotlib.animation as animation
import numpy as np

def plotDataset(x, dataset, index=0, n=9):
    '''
    Function to visualize MNIST data set
    Parameters
    ----------
    x : multi-dimensional array
    index : integer
            plot figure i
    n : integer
        plot n figures 
    Returns
    -------
    None.
    '''
    num = range(index, index+n)
    counter=0
    if dataset is 'CIFAR10':
        for i in num:
            # define subplot
            plt.subplot(330 + 1 + counter)
            # reshape array and define rgb values
            im_r = x[i][0:1024].reshape(32, 32)
            im_g = x[i][1024:2048].reshape(32, 32)
            im_b = x[i][2048:].reshape(32, 32)
            # reconnect rgb arrays
            img = np.dstack((im_r, im_g, im_b))   
            # plot raw pixel data
            plt.imshow(img)
            # increment counter to draw multiple images in same plot
            counter = counter+1
    else:
        for i in num:
            # define subplot
            plt.subplot(330 + 1 + counter)
            im1 = x[i][0:784].reshape(28,28)
            # reshape image and plot raw pixel data
            plt.imshow(x[i].reshape(28,28), cmap=plt.get_cmap('gray'))
            # increment counter to draw multiple images in same plot
            counter = counter+1
    # show the figure
    plt.show()
    

    
def visualizeMetrics(metrics_output, ep):
    epochs = range(1,ep+1,1)
    loss_train = []
    acc_train =[]
    f1_train=[]
    p_train=[]
    r_train=[]
    
    for key,val in metrics_output.items():
        if key.startswith("loss"):
            loss_train.append(val)
        elif key.startswith("accuracy"):
            acc_train.append(val)
        elif key.startswith("f1"):
            f1_train.append(val)
        elif key.startswith("precision"):
            p_train.append(val)
        elif key.startswith("recall"):
            r_train.append(val)
    
    if len(loss_train) == ep:
        plt.plot(epochs, loss_train, label='Training Loss')
        plt.title('Training loss')
        plt.xlabel('Epochs')
        plt.ylabel('Loss')
        plt.show()
    if len(acc_train) == ep:
        plt.plot(epochs, acc_train, label='Training Accuracy')
        plt.title('Training Accuracy')
        plt.xlabel('Epochs')
        plt.ylabel('Accuracy')
        plt.show()
    if len(f1_train) == ep:
        plt.plot(epochs, f1_train, label='Training F1 score')
        plt.title('Training F1 score')
        plt.xlabel('Epochs')
        plt.ylabel('F1 score')
        plt.show()
    if len(p_train) == ep:
        plt.plot(epochs, p_train, label='Training Precision')
        plt.title('Training Precision')
        plt.xlabel('Epochs')
        plt.ylabel('Precision')
        plt.show()
    if len(r_train) == ep:
        plt.plot(epochs, r_train, label='Training Recall')
        plt.title('Training Recall')
        plt.xlabel('Epochs')
        plt.ylabel('Recall')
        plt.show()



import numpy as np
from sklearn.metrics import confusion_matrix, f1_score, accuracy_score, precision_score, recall_score
import pandas as pd


# def absolute_metrics(y_hat,y_label):
#     abs_value=0
#     i=0
#     while i<len(y_hat):
#         abs_value= abs_value+(abs(y_hat[i]-y_label[i]))
#         i=i+1
#     abs_value = abs_value / len(y_hat)
#     return abs_value

def accuracy_metrics(y_hat,y_label):
    count=0
    i=0
    while i<len(y_hat):
        if y_hat[i]==y_label[i] :
            count=count+1
        i=i+1
    accur=count/len(y_hat)
    return accur

def confusion_matrix(y_hat, y_label):

    classes = np.unique(y_label) # extract the different classes
    matrix = np.zeros((len(classes), len(classes))) # initialize the confusion matrix with zeros

    for i in range(len(classes)):
        for j in range(len(classes)):

            matrix[i, j] = np.sum((y_label == classes[i]) & (y_hat == classes[j]))

    return matrix


def TP(y_hat, y_label,Type):
    conf_mat = confusion_matrix(y_hat, y_label)
    if Type == 'micro':
        return sum(np.diag(conf_mat))
    elif Type == 'macro':
        return np.diag(conf_mat)

def FP(y_hat, y_label,Type):
    if Type == 'micro':
        fp=[]
        conf_mat = confusion_matrix(y_hat, y_label)
        for i in range(conf_mat.shape[1]):
            fp.append(sum(conf_mat[:,i]) - conf_mat[i,i])
        return sum(fp)
    elif Type == 'macro':
        conf_mat = confusion_matrix(y_hat, y_label)
        return np.sum(conf_mat, axis=0) - TP(y_hat, y_label,'macro')
        

def FN(y_hat, y_label , Type):
    if Type == 'micro':
        fn=[]
        conf_mat = confusion_matrix(y_hat, y_label)
        for i in range(conf_mat.shape[1]):
            fn.append(sum(conf_mat[i,:]) - conf_mat[i,i])
        return sum(fn)
    elif Type == 'macro':
        conf_mat = confusion_matrix(y_hat, y_label)
        return np.sum(conf_mat, axis=1) - TP(y_hat, y_label,'macro')
        

def Accuracy(y_hat, y_label, data_size):
    return np.sum(TP(y_hat, y_label)/data_size)

def Recall(y_hat, y_label,Type):
    if Type == 'micro':    
        tp = TP(y_hat, y_label,'micro')
        fp = FP(y_hat, y_label,'micro')
        return tp / (tp+fp)
    
    elif Type == 'macro':
        return np.mean(TP(y_hat, y_label,'macro') / (TP(y_hat, y_label,'macro') + FN(y_hat, y_label,'macro')))
        
    

def Precision(y_hat, y_label,Type):
    if Type == 'micro':       
        tp = TP(y_hat, y_label,'micro')
        fn = FN(y_hat, y_label,'micro')
        return tp / (tp+fn)
    elif Type == 'macro':
        return np.mean(TP(y_hat, y_label,'macro') / (TP(y_hat, y_label,'macro') + FP(y_hat, y_label,'macro')))

        
        
def F1_score(y_hat, y_label,Type):
    if TP(y_hat, y_label,'micro') > 0:
        P = Precision(y_hat, y_label,Type)
        R = Recall(y_hat, y_label,Type)
        return 2*((P*R)/(P+R)) 
    else:
        return 0


# a = confusion_matrix(trainLabels_.T,Y_hat.T)
# TP(Y_hat.T,trainLabels_.T,'macro')
# FP(Y_hat.T,trainLabels_.T,'macro')
# FN(Y_hat.T,trainLabels_.T,'macro')
# Precision(Y_hat.T,trainLabels_.T,'micro')
# Recall(Y_hat.T,trainLabels_.T,'micro')
# F1_score(Y_hat.T,trainLabels_.T,'macro')


# print (precision_score(y_class.T,pred.T,average='micro'))
# print (recall_score(y_class.T, pred.T, average='micro'))
# print (f1_score(y_class.T, pred.T, average='micro'))
# print(accuracy_score(y_class.T,predict.T))



import numpy as np
from Terminatetensorflow.actvfn import *

class layer:
    def __init__(self, output_dims, input_dims, act_fn):
        self.weights = np.random.randn(output_dims, input_dims) * np.sqrt(2/input_dims)
        self.biases = np.zeros((output_dims,1))
        self.act_fn = act_fn
        self.Z = None
        self.A = None     
        self.dW = np.zeros((output_dims, input_dims))
        self.db = np.zeros((output_dims,1))
        self.dZ = None
        self.dA = None
        self.v_w = np.zeros((output_dims, input_dims))
        self.v_b = np.zeros((output_dims,1))
        self.s_w = np.zeros((output_dims, input_dims))
        self.s_b = np.zeros((output_dims,1))
        
        
    def forward_prop(self, input):
        self.Z = np.dot(self.weights, input) + self.biases
        
        
        if self.act_fn == "sigmoid":
            self.A = sigmoid(self.Z)
        if self.act_fn == "relu":
            self.A = relu(self.Z)
        if self.act_fn == "tanh":
            self.A = tanh(self.Z)
        if self.act_fn == "linear":
            self.A = linear(self.Z)
        if self.act_fn == "softmax":
            self.A = softmax(self.Z)
            
        return self.A
        
    def backward_prop(self, prev_A, dA_prev, y, output_layer = False):
        
        if output_layer is True:
            self.dZ = self.A - y            
            self.dW = (1./y.shape[1]) * np.dot(self.dZ, prev_A.T)
            self.db = (1./y.shape[1]) * np.sum(self.dZ, axis=1, keepdims=True)
            dA_prev = np.dot(self.weights.T, self.dZ)
               
        else:
            if self.act_fn == "sigmoid":
                self.dZ = dA_prev * derivative_sigmoid(self.Z)
            elif self.act_fn == "tanh":
                self.dZ = dA_prev * derivative_tanh(self.Z)
            elif self.act_fn == "relu":
                self.dZ = dA_prev * d_relu(self.Z)
            elif self.act_fn == "linear":
                self.dZ = dA_prev * d_linear(self.Z)
                
            self.dW = (1./y.shape[1]) * np.dot(self.dZ, prev_A.T)
            self.db = (1./y.shape[1]) * np.sum(self.dZ, axis=1, keepdims=True)
            dA_prev = np.dot(self.weights.T, self.dZ)
            
        return dA_prev, self.dW, self.db
    
    def update_parameters(self, learning_rate):
        self.weights = self.weights -  learning_rate * self.dW
        self.biases= self.biases -  learning_rate * self.db

    def update_parameters_momentum(self, learning_rate, gamma=0.9):
        self.v_w = gamma*self.v_w +  (1 - gamma)* self.dW
        self.v_b = gamma*self.v_b +  (1 - gamma)* self.db
        self.weights = self.weights - learning_rate * self.v_w
        self.biases = self.biases -  learning_rate  *  self.v_b
        
    def update_parameters_rmsprop(self, learning_rate, beta=0.9, epsilon = 1e-8):
        self.s_w = beta*self.s_w +  (1 - beta)* np.power(self.dW,2)
        self.s_b = beta*self.s_b +  (1 - beta)* np.power(self.db,2)
        self.weights = self.weights - learning_rate * np.divide(self.dW,np.sqrt(self.s_w + epsilon))
        self.biases = self.biases -  learning_rate  *  np.divide(self.db,np.sqrt(self.s_b + epsilon))

        
    def update_parameters_adam(self, learning_rate, t = 2, beta1 = 0.9, beta2 = 0.999,  epsilon = 1e-8):
        self.v_w = beta1*self.v_w +  (1 - beta1)* self.dW
        self.v_b = beta1*self.v_b +  (1 - beta1)* self.db
        v_wcorrected = self.v_w / (1-np.power(beta1,t))
        v_bcorrected = self.v_b / (1-np.power(beta1,t))
        self.s_w = beta2*self.s_w + (1-beta2)*np.power(self.dW,2)
        self.s_b = beta2*self.s_b + (1-beta2)*np.power(self.db,2)
        s_wcorrected = self.s_w / (1-np.power(beta2,t))
        s_bcorrected = self.s_b / (1-np.power(beta2,t))
        self.weights = self.weights - learning_rate*v_wcorrected / np.sqrt(s_wcorrected + epsilon)
        self.biases = self.biases - learning_rate*v_bcorrected / np.sqrt(s_bcorrected + epsilon)




import numpy as np
import math

def sigmoid(x):
   # np.clip(x,np.finfo('int64').max,-np.finfo('int64').max)
    s = 1/(1+np.exp(-x)) #np.exp
    return s


def tanh(x):
   # t = (np.exp(x)-np.exp(-x))/(np.exp(x)+np.exp(-x))
    t = 1.7159 * np.tanh(x)
    return t

def linear(x):
    return x


def relu(x):
    return np.maximum(0.0, x)


def softmax(x):
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum(axis=0)

def derivative_sigmoid(x):
    return (sigmoid(x) * (1- sigmoid(x)))


def derivative_tanh(x):
    f = 1-(tanh(x)*tanh(x))
    return f


def d_linear(func):
    return np.ones(func.shape)


def d_relu(x):
    return np.greater(x, 0).astype(int)  



def d_softmax(X):
    x = softmax(X)
    s = x.reshape(-1 , 1)
    return (np.diagflat(s) - np.dot(s, np.transpose(s)))


